unit ByteArrayHandler;

interface

uses
    Windows, StrUtils, SysUtils, Classes, WMapConst, TypesDefs, Winsock2, SyncObjs, LibeayFunc;

const
     SIZE32     = $20;
     SIZE64     = $40;
     SIZE128    = $80;
     SIZE256    = $100;
     SIZE512    = $200;
     SIZE1K     = $400;
     SIZE2K     = $800;
     SIZE4K     = $1000;
     SIZE8K     = $2000;
     SIZE16K    = $4000;
     SIZE32K    = $8000;
     SIZE64K    = $10000;
     SIZE128K   = $20000;
     SIZE256K   = $40000;
     SIZE512K   = $80000;

     CMSG_HDR_SIZE  = 6;
     SMSG_HDR_SIZE  = 4;
     CRYPT_HASH_SIZE = 40;

type
  PHashArray = ^THashArray;
  THashArray = array[0..CRYPT_HASH_SIZE - 1]of byte;

  TSS_Hash = packed record
    Hash: THashArray;
    temp,arr_pos: byte;
    FirstPacket: boolean;
  end;

  TSendSession = class;

  TSendBuffer = class
  protected
    FCS: TCriticalSection;
    FData: TBytes;
    FARRAY_SIZE, FARRAY_MASK: dword;
    FTail, FHead: dword;
    FBusy: boolean;
    function GetDataCount:dword;
    function IsOverflow(add_size: dword): boolean;
    function GetToSend: dword;
    function GetDataPtr: pointer;
  public
    property data: TBytes read FData;
    property Tail: dword read FTail;
    property Head: dword read FHead write FHead;
    constructor Create(size: dword);virtual;
    destructor Destroy;override;
    function AddData(buf: pointer; len: dword): boolean;
    procedure Send(csock: TObject; sock: dword);
    procedure IncSent(sent: dword);
  end;

  TReadBuffer = class
  private
    FData: TBytes;
    //FARR_SIZE, FARR_MASK: dword;
    FTail, FHead: dword;
  public
    constructor Create; virtual;
    destructor Destroy; override;
    function GetDataCount: dword;
    function IncReceived(add_size: dword): boolean;
    function GetDataPtr: pointer;
    function GetHeadPtr: pointer;
    function GetBuffFreeSpace: dword;
    procedure ReleaseBufferCount(done: dword);
  end;

  TSendSession = class
  private
    FData: TBytes;
    FOffs: dword;
    FOpCode: word;
    procedure SetOpcode(OpCode: word);
    //procedure SetSize;
  public
    property data: TBytes read FData write FData;
    property Offs: dword read FOffs;
    property OpCode: word read FOpCode;
    constructor Create(arr_size: dword = MaxByte);overload;
    constructor Create(const fromSession: TSendSession);overload;
    destructor Destroy; override;
    procedure Init(OpCode: word);
    procedure InitAuth(OpCode: byte);
    //procedure Finalize;
    procedure Add(const in_param: byte); overload;
    procedure Add(const in_param: word); overload;
    procedure Add(const in_param: integer); overload;
    procedure Add(const in_param: dword); overload;
    procedure Add(const in_param: int64); overload;
    procedure Add(const in_param: uint64); overload;
    procedure Add(const in_param: single); overload;
    procedure Add(const in_param: double); overload;
    procedure Add(buf: pointer; add_size: dword); overload;
    procedure Add(const in_param: string); overload;
    procedure Add(const in_param: TPos); overload;
    procedure Add(const in_param: TPosR); overload;
    procedure AddPackedGUID(const in_param: int64); overload;
    procedure AddPackedGUID(const in_param: uint64); overload;
    procedure PutToPos(const in_param: byte; pos_offs: integer); overload;
    procedure PutToPos(const in_param: word; pos_offs: integer); overload;
    procedure PutToPos(const in_param: integer; pos_offs: integer); overload;
    procedure PutToPos(const in_param: dword; pos_offs: integer); overload;
    procedure PutToPos(const in_param: int64; pos_offs: integer); overload;
    procedure PutToPos(const in_param: uint64; pos_offs: integer); overload;
    procedure PutToPos(const in_param: single; pos_offs: integer); overload;
    procedure PutToPos(const in_param: double; pos_offs: integer); overload;
  end;

  function SplitStr(const InStr,Delim: string):TStringArray;

implementation

uses IOCPHandler, LogHandler;

//const
//    SMSG_key: array[0..SEED_KEY_SIZE-1]of byte = ( $22, $BE, $E5, $CF, $BB, $07, $64, $D9, $00, $45, $1B, $D0, $24, $B8, $D5, $45 );
//    CMSG_key: array[0..SEED_KEY_SIZE-1]of byte = ( $F4, $66, $31, $59, $FC, $83, $6E, $31, $31, $02, $51, $D5, $44, $31, $67, $98 );

function bswap(in_data: dword):dword;
asm
   bswap eax
end;

function SplitStr(const InStr,Delim: string):TStringArray;
var
  offs,prevOffs,count: integer;
begin
    prevOffs:= 1;
    result:= nil;
    while true do begin
      offs:= posEx(Delim, InStr, prevOffs);
      if offs = 0 then begin
        if prevOffs <= length(InStr) then begin
          count:= length(result);
          setlength(result, count + 1);
          result[count]:= Copy(InStr, prevOffs, length(InStr) - prevOffs + 1);
        end;
        break;
      end;
      count:= length(result);
      setlength(result, count + 1);
      result[count]:= Copy(InStr, prevOffs, offs - prevOffs);
      prevOffs:= offs + 1;
    end;
end;

{ TSendBuffer }

constructor TSendBuffer.Create(size: dword);
begin
    FCS:= TCriticalSection.Create;
    FTail:= 0;
    FHead:= 0;
    FBusy:= false;
    FARRAY_SIZE:= size;
    FARRAY_MASK:= FARRAY_SIZE - 1;
    setlength(FData, FARRAY_SIZE);
end;

destructor TSendBuffer.Destroy;
begin
    FData:= nil;
    FCS.Free;
    inherited;
end;

function TSendBuffer.IsOverflow(add_size: dword):boolean;
begin
    result:= GetDataCount + add_size >= FARRAY_SIZE;
end;

function TSendBuffer.AddData(buf: pointer; len: dword): boolean;
var
   expected_end,up_to_end: dword;
begin
    result:= not(IsOverflow(len));
    if not(result) then begin
      //AddToLog('IsOverflow SendLoopBuffer datalen='+IntToStr(GetDataCount)+' +data='+IntToStr(len)+' FHead='+IntToStr(FHead)+' FTail='+IntToStr(FTail));
      exit;
    end;
    FCS.Enter;
    expected_end:= (FHead + len) and FARRAY_MASK;
    if expected_end >= FHead then
      move(buf^, FData[FHead], len)
    else begin
      up_to_end:= FARRAY_SIZE - FHead;
      move(buf^, FData[FHead], up_to_end);
      move(TBytes(buf)[up_to_end], FData[0], len - up_to_end);
    end;
    FHead:= expected_end;
    result:= true;
    FCS.Leave;
end;

function TSendBuffer.GetDataCount:dword;
begin
    if FTail <= FHead then result:= FHead - FTail //linear
    else result:= FARRAY_SIZE - FTail + FHead;
end;

function TSendBuffer.GetToSend: dword;
begin
     if FTail <= FHead then result:= FHead - FTail
     else result:= FARRAY_SIZE - FTail;
end;

function TSendBuffer.GetDataPtr: pointer;
begin
    result:= @FData[FTail];
end;

procedure TSendBuffer.Send(csock: TObject; sock: dword);
begin
    if FBusy or (GetToSend = 0) then exit;
    FBusy:= true;
    TCustomSocket(csock).Send(sock, GetDataPtr, GetToSend);
end;

procedure TSendBuffer.IncSent(sent: dword);
begin
    FTail:= (FTail + sent)and FARRAY_MASK;
    FBusy:= false;
end;

{ TReadBuffer }

constructor TReadBuffer.Create;
begin
    setlength(FData,SIZE4K);
    FHead:= 0;
    FTail:= 0;
end;

destructor TReadBuffer.Destroy;
begin
    FData:= nil;
end;

procedure TReadBuffer.ReleaseBufferCount(done: dword);
begin
    Assert(FTail + done < dword(length(FData)),'TReadBuffer.GetPacket read out of data');
    inc(FTail, done);
    if (FHead <> FTail)or(FHead = 0) then exit;
    FHead:= 0;
    FTail:= 0;
    if length(FData) > SIZE4K then setlength(FData, SIZE4K);
end;

function TReadBuffer.GetBuffFreeSpace: dword;
begin
    Assert(FHead < dword(length(FData)), 'TReadBuffer.GetBuffFreeSpace: FHead >= dword(length(FData))');
    result:= dword(high(FData)) - FHead;
    if result < SIZE1K then begin
      setlength(FData, ((length(FData) shr 12) + 1) shl 12); //inc by 4K
      result:= dword(high(FData)) - FHead;
    end;
end;

function TReadBuffer.GetDataCount: dword;
begin
    Assert(FTail <= FHead,'TReadBuffer FTail > FHead');
    result:= FHead - FTail;
end;

{function TReadBuffer.AddPacket(buf: pointer; add_size: dword): boolean;
var
  explen: dword;
begin
    explen:=FHead + add_size;
    result:=explen < SIZE8K;
    if explen > dword(length(FData)) then
      setlength(FData,((explen shr 12) + 1) shl 12); //inc by 4K
    move(buf^,FData[FHead],add_size);
    FHead:=explen;
end;}

function TReadBuffer.IncReceived(add_size: dword): boolean;
var
  explen: dword;
begin
    explen:= FHead + add_size;
    result:= true; //explen < SIZE8K; //we dont want to get huge packets
    if explen > dword(length(FData)) then
      setlength(FData, ((explen shr 12) + 1) shl 12); //inc by 4K
    FHead:= explen;
end;

function TReadBuffer.GetDataPtr: pointer;
begin
    result:= @FData[FTail];
end;

function TReadBuffer.GetHeadPtr: pointer;
begin
    result:= @FData[FHead];
end;

{ TSendSession }

constructor TSendSession.Create(arr_size: dword = MaxByte);
begin
     setlength(FData,arr_size);
     FOffs:=0;
end;

constructor TSendSession.Create(const fromSession: TSendSession);
begin
     setlength(FData,fromSession.Offs);
     move(fromSession.data[0],FData[0],fromSession.Offs);
     FOpCode:=fromSession.OpCode;
     FOffs:=fromSession.Offs;
end;

destructor TSendSession.Destroy;
begin
     FData:=nil;
     inherited;
end;

procedure TSendSession.SetOpcode(OpCode: word);
begin
     FOpCode:=OpCode;
	   FData[2]:= OpCode;
	   FData[3]:= OpCode shr 8;
end;

{procedure TSendSession.SetSize;
begin
     FData[0]:= (FOffs - 2) shr 8;
     FData[1]:= FOffs - 2;
end;}

procedure TSendSession.Init(OpCode: word);
begin
		 SetOpcode(OpCode);
     FOffs:=SMSG_HDR_SIZE;
end;

procedure TSendSession.InitAuth(OpCode: byte);
begin
     ZeroMemory(FData,length(FData));
	   FData[0]:= OpCode;
     FOffs:=1;
end;

{procedure TSendSession.Finalize;
var
   mstream: TFileStream;
   fname: string;
begin
     SetSize;

     {fname:=AppPath+'\dump'+IntToStr(sckt)+'.s';
     if not(FileExists(fname)) then
        mstream:= TFileStream.Create(fname,fmCreate)
     else mstream:= TFileStream.Create(fname,fmOpenReadWrite);
     mstream.Seek(mstream.Size,soFromBeginning);
     mstream.Write(data[0],Offs);
     mstream.Free;
end;}

procedure TSendSession.Add(const in_param: byte);
asm
   mov ecx,[eax+FData]
   add ecx,[eax+FOffs]
   mov [ecx],dl
   inc dword ptr [eax+FOffs]
{begin
     FData[FOffs]:=in_param;
     inc(FOffs);}
end;

procedure TSendSession.Add(const in_param: word);
asm
   mov ecx,[eax+FData]
   add ecx,[eax+FOffs]
   mov [ecx],dx
   add dword ptr [eax+FOffs],2
{begin
     move(in_param,FData[FOffs],sizeof(word));
     inc(FOffs,sizeof(word));}
end;

procedure TSendSession.Add(const in_param: integer);
asm
   mov ecx,[eax+FData]
   add ecx,[eax+FOffs]
   mov [ecx],edx
   add dword ptr [eax+FOffs],4
{begin
     move(in_param,FData[FOffs],sizeof(integer));
     inc(FOffs,sizeof(integer));}
end;

procedure TSendSession.Add(const in_param: dword);
asm
   mov ecx,[eax+FData]
   add ecx,[eax+FOffs]
   mov [ecx],edx
   add dword ptr [eax+FOffs],4
{begin
     move(in_param,FData[FOffs],sizeof(dword));
     inc(FOffs,sizeof(dword));}
end;

procedure TSendSession.Add(const in_param: int64);
asm
   mov ecx,[eax+FData]
   add ecx,[eax+FOffs]
   mov edx,[ebp+8]
   mov [ecx],edx
   mov edx,[ebp+$C]
   mov [ecx+4],edx
   add dword ptr [eax+FOffs],8
{begin
     move(in_param,FData[FOffs],sizeof(int64));
     inc(FOffs,sizeof(int64));}
end;

procedure TSendSession.Add(const in_param: uint64);
asm
   mov ecx,[eax+FData]
   add ecx,[eax+FOffs]
   mov edx,[ebp+8]
   mov [ecx],edx
   mov edx,[ebp+$C]
   mov [ecx+4],edx
   add dword ptr [eax+FOffs],8
{begin
     move(in_param,FData[FOffs],sizeof(uint64));
     inc(FOffs,sizeof(uint64));}
end;

procedure TSendSession.Add(const in_param: single);
asm
   mov ecx,[eax+FData]
   add ecx,[eax+FOffs]
   mov edx,[ebp+8]
   mov [ecx],edx
   add dword ptr [eax+FOffs],4
{begin
     move(in_param,FData[FOffs],sizeof(single));
     inc(FOffs,sizeof(single));}
end;

procedure TSendSession.Add(const in_param: double);
asm
   mov ecx,[eax+FData]
   add ecx,[eax+FOffs]
   mov edx,[ebp+8]
   mov [ecx],edx
   mov edx,[ebp+$C]
   mov [ecx+4],edx
   add dword ptr [eax+FOffs],8
{begin
     move(in_param,FData[FOffs],sizeof(double));
     inc(FOffs,sizeof(double));}
end;

procedure TSendSession.Add(const in_param: string);
var
   str_len: integer;
begin
     str_len:=length(in_param);
     if str_len=0 then begin
        FData[FOffs]:=0;inc(FOffs);
     end
     else begin
          inc(str_len);
          move(in_param[1],FData[FOffs],str_len);
          inc(FOffs,str_len);
     end;
end;

procedure TSendSession.Add(const in_param: TPos);
begin
     move(in_param,FData[FOffs],sizeof(TPos));
     inc(FOffs,sizeof(TPos));
end;

procedure TSendSession.Add(const in_param: TPosR);
begin
     move(in_param,FData[FOffs],sizeof(TPosR));
     inc(FOffs,sizeof(TPosR));
end;

procedure TSendSession.Add(buf: pointer; add_size: dword);
begin
     move(buf^,FData[FOffs],add_size);
     inc(FOffs,add_size);
end;

procedure TSendSession.AddPackedGUID(const in_param: int64);
var
   mask_offs: integer;
   mask: byte;
begin
     mask_offs:=FOffs;inc(FOffs);mask:=0;
     if TBytes(@in_param)[0] <> 0 then begin FData[FOffs]:=TBytes(@in_param)[0]; inc(FOffs); mask:=mask or 1; end;
     if TBytes(@in_param)[1] <> 0 then begin FData[FOffs]:=TBytes(@in_param)[1]; inc(FOffs); mask:=mask or 2; end;
     if TBytes(@in_param)[2] <> 0 then begin FData[FOffs]:=TBytes(@in_param)[2]; inc(FOffs); mask:=mask or 4; end;
     if TBytes(@in_param)[3] <> 0 then begin FData[FOffs]:=TBytes(@in_param)[3]; inc(FOffs); mask:=mask or 8; end;
     if TBytes(@in_param)[4] <> 0 then begin FData[FOffs]:=TBytes(@in_param)[4]; inc(FOffs); mask:=mask or $10; end;
     if TBytes(@in_param)[5] <> 0 then begin FData[FOffs]:=TBytes(@in_param)[5]; inc(FOffs); mask:=mask or $20; end;
     if TBytes(@in_param)[6] <> 0 then begin FData[FOffs]:=TBytes(@in_param)[6]; inc(FOffs); mask:=mask or $40; end;
     if TBytes(@in_param)[7] <> 0 then begin FData[FOffs]:=TBytes(@in_param)[7]; inc(FOffs); mask:=mask or $80; end;
     FData[mask_offs]:=mask;
end;

procedure TSendSession.AddPackedGUID(const in_param: uint64);
var
   mask_offs: integer;
   mask: byte;
begin
     mask_offs:=FOffs;inc(FOffs);mask:=0;
     if TBytes(@in_param)[0] <> 0 then begin FData[FOffs]:=TBytes(@in_param)[0]; inc(FOffs); mask:=mask or 1; end;
     if TBytes(@in_param)[1] <> 0 then begin FData[FOffs]:=TBytes(@in_param)[1]; inc(FOffs); mask:=mask or 2; end;
     if TBytes(@in_param)[2] <> 0 then begin FData[FOffs]:=TBytes(@in_param)[2]; inc(FOffs); mask:=mask or 4; end;
     if TBytes(@in_param)[3] <> 0 then begin FData[FOffs]:=TBytes(@in_param)[3]; inc(FOffs); mask:=mask or 8; end;
     if TBytes(@in_param)[4] <> 0 then begin FData[FOffs]:=TBytes(@in_param)[4]; inc(FOffs); mask:=mask or $10; end;
     if TBytes(@in_param)[5] <> 0 then begin FData[FOffs]:=TBytes(@in_param)[5]; inc(FOffs); mask:=mask or $20; end;
     if TBytes(@in_param)[6] <> 0 then begin FData[FOffs]:=TBytes(@in_param)[6]; inc(FOffs); mask:=mask or $40; end;
     if TBytes(@in_param)[7] <> 0 then begin FData[FOffs]:=TBytes(@in_param)[7]; inc(FOffs); mask:=mask or $80; end;
     FData[mask_offs]:=mask;
end;

procedure TSendSession.PutToPos(const in_param: byte; pos_offs: integer);
asm
   mov eax,[eax+FData]
   mov [eax+ecx],dl
{begin
     FData[pos_offs]:=in_param;  //+SMSG_HDR_SIZE}
end;

procedure TSendSession.PutToPos(const in_param: word; pos_offs: integer);
asm
   mov eax,[eax+FData]
   mov [eax+ecx],dx
{begin
     move(in_param,FData[pos_offs],sizeof(word));}
end;

procedure TSendSession.PutToPos(const in_param: integer; pos_offs: integer);
asm
   mov eax,[eax+FData]
   mov [eax+ecx],edx
{begin
     move(in_param,FData[pos_offs],sizeof(integer));}
end;

procedure TSendSession.PutToPos(const in_param: dword; pos_offs: integer);
asm
   mov eax,[eax+FData]
   mov [eax+ecx],edx
{begin
     move(in_param,FData[pos_offs],sizeof(dword));}
end;

procedure TSendSession.PutToPos(const in_param: int64; pos_offs: integer);
asm
   mov eax,[eax+FData]
   add eax, ecx
   mov ecx,[ebp+8]
   mov [eax],ecx
   mov ecx,[ebp+$C]
   mov [eax+4],ecx
{begin
     move(in_param,FData[pos_offs],sizeof(int64));}
end;

procedure TSendSession.PutToPos(const in_param: uint64; pos_offs: integer);
asm
   mov eax,[eax+FData]
   add eax, ecx
   mov ecx,[edx]
   mov [eax],ecx
   mov ecx,[edx+4]
   mov [eax+4],ecx
{begin
     move(in_param,FData[pos_offs],sizeof(uint64));}
end;

procedure TSendSession.PutToPos(const in_param: single; pos_offs: integer);
asm
   mov eax,[eax+FData]
   mov [eax+ecx],edx
{begin
     move(in_param,FData[pos_offs],sizeof(single));}
end;

procedure TSendSession.PutToPos(const in_param: double; pos_offs: integer);
asm
   mov eax,[eax+FData]
   add eax, ecx
   mov ecx,[edx]
   mov [eax],ecx
   mov ecx,[edx+4]
   mov [eax+4],ecx
{begin
     move(in_param,FData[pos_offs],sizeof(double));}
end;

end.
