unit LoginHandler;

interface

uses
  Windows, SysUtils, Classes, WinSock2, TypesDefs, SyncObjs, AcedContainers,
  UnitMain, IocpHandler, ByteArrayHandler, 
  libeay32, LibeayFunc, PostgresClasses, libpq_fe;

const
  MAX_USERNAME_LEN    = 50;
  MAX_PASSWORD_LEN    = 50;
  MAX_CHARACTER_LEN   = 50;

type
  AuthProcs = function(pdata: pointer; count: dword; sckt: TSocket):dword;

  PCMSG_CHALLENGE = ^TCMSG_CHALLENGE;
  TCMSG_CHALLENGE = packed record
    OpCode: byte;                           //0
    ErrorCode: byte;                        //1
    Size: word;                             //2
    ClientType: array [0..3] of char;       //4
    Version: array [0..2] of byte;          //8
    Build: word;                            //11
    CPU: array [0..3] of char;              //13
    OS: array [0..3] of char;               //17
    Language: array [0..3] of char;         //21
    TimeZone: dword;                        //25
    IP: dword;                              //29
    LoginSize: byte;                        //33
    Login: array[0..0]of char;              //34
  end;

  {PSMSG_CHALLENGE = ^TSMSG_CHALLENGE;
  TSMSG_CHALLENGE = packed record
    OpCode: Byte;
    UNK1: Byte; // ??? Grunt command
    ErrorCode: Byte;
    SRP6_B: array [0..SIZE32 - 1] of Byte;
    SRP6_g_Length: Byte;
    SRP6_g: array [0..0] of Byte;
    SRP6_N_Length: Byte;
    SRP6_N: array [0..SIZE32 - 1] of Byte;
    SRP6_Salt: array [0..SIZE32 - 1] of Byte;
    Seed: array [0..15] of Byte;
    Security: Byte; // Possible values: 0 - None, 1 - PIN, 2 - Matrix, 4 - Require Security Token
  end;}

  PCMSG_PROOF = ^TCMSG_PROOF;
  TCMSG_PROOF = packed record
    OpCode: Byte;
    A: array [0..SIZE32 - 1] of byte;
    M1: array [0..SHA_DIGEST_LENGTH - 1] of byte;
    mdctx: array [0..SHA_DIGEST_LENGTH - 1] of byte;
    KeyCount: byte;
    Security: byte;
  end;

  {PSMSG_PROOF = ^TSMSG_PROOF;
  TSMSG_PROOF = packed record
    OpCode: Byte;
    ErrorCode: Byte;
    SRP6_M2: array [0..SRP6_M2_LENGTH - 1] of Byte;
    UNK1: Cardinal;
    UNK2: Cardinal;
    UNK3: Word;
  end;}

  {PCMSG_RECHALLENGE = ^TCMSG_RECHALLENGE;
  TCMSG_RECHALLENGE = TCMSG_CHALLENGE;

  PSMSG_RECHALLENGE = ^TSMSG_RECHALLENGE;
  TSMSG_RECHALLENGE = packed record
    OpCode: Byte;
    ErrorCode: Byte;
    Seed: array [0..15] of Byte;
    Zero: array [0..15] of Byte;
  end;

  PCMSG_REPROOF = ^TCMSG_REPROOF;
  TCMSG_REPROOF = packed record
    OpCode: Byte;
    Seed: array [0..15] of Byte;
    mdctx: array [0..19] of Byte;
    UNK1: array [0..19] of Byte; // crc_hash
    UNK2: Byte; // number of keys
  end;}

  {PSMSG_REPROOF = ^TSMSG_REPROOF;
  TSMSG_REPROOF = packed record
    OpCode: Byte;
    ErrorCode: Byte;
    Zero: Word;
  end;}

  PCMSG_REALMLIST = ^TCMSG_REALMLIST;
  TCMSG_REALMLIST = packed record
    OpCode: Byte;
    UNK1: Cardinal;
  end;

  {PSMSG_REALMLIST = ^TSMSG_REALMLIST;
  TSMSG_REALMLIST = packed record
    OpCode: Byte;
    Size: Word;
    UNK1: Integer;
    Count: Word;
  end;

  PSMSG_REALMLIST_ENTRY_BEGIN = ^TSMSG_REALMLIST_ENTRY_BEGIN;
  TSMSG_REALMLIST_ENTRY_BEGIN = packed record
    Rules: Byte
  20:07:41: ;
    Lock: Byte;
    Status: Byte;
  end;

  PSMSG_REALMLIST_ENTRY_END = ^TSMSG_REALMLIST_ENTRY_END;
  TSMSG_REALMLIST_ENTRY_END = packed record
    Load: Single;
    Characters: Byte;
    Locale: Byte;
    UNK1: Byte;
  end;}

  PFieldData = ^TFieldData;
  TFieldData = record
    field_name: string;
    idx: integer;
    pdata: TBytes;
    data_len: dword;
    is_null: boolean;
  end;

  TFieldDataArray = array of TFieldData; 

  TAuthPlayer = class(TObject)
  public
    id: dword;
    username: UTF8String;
    is_active, banned, IsAuthed: boolean;
    gm_level: integer;
    client_version, timezone: dword;
    expires: TDateTime;
    Hash: array[0..CRYPT_HASH_SIZE - 1]of byte;
    salt: array[0..SIZE32-1]of byte;
    biB, biN, biV, biChRes: pBIGNUM;
    //Socket: TSocket;
    constructor Create; //(sock: TSocket)
    destructor Destroy;override;
  end;

  TThreadedAuthPlayer = class(TObject)
  private
    FLock: TRTLCriticalSection;
    FObject: TAuthPlayer;
  public
    constructor Create; //(sock: TSocket)
    destructor Destroy;override;
    function Lock: TAuthPlayer;
    procedure UnLock;
  end;

  function StoreUsernamePassHash(const username, password: UTF8String; gm_level: integer): integer;
  procedure OnReadAuthSocket(ppsContext: PPER_SOCKET_CONTEXT; count: dword);
  procedure OnWriteAuthSocket(ppsContext: PPER_SOCKET_CONTEXT; count: dword);
  procedure OnAcceptAuthSocket(ppsContext: PPER_SOCKET_CONTEXT);
  procedure OnDisconnectAuthSocket(ppsContext: PPER_SOCKET_CONTEXT);

implementation

uses RealmsHandler, LogHandler, QueryThreads;

const
    BIGNUMBER_N_STR                 = '894B645E89E1535BBDAD5B8B290650530801B18EBFBF5E8FAB3C82872A3E9BB7';
    BIGNUMBER_B_STR                 = '8692E3A6BA48B5B1004CEF76825127B7EB7D1AEF';

    CE_SUCCESS                      = $00;
	  CE_IPBAN                        = $01; //2bd -- unable to connect (some internal problem)
    CE_ACCOUNT_CLOSED               = $03; // "This account has been closed and is no longer in service -- Please check the registered email address of this account for further information.";
    CE_ACCOUNT_NO_MATCH             = $04; //(5)The information you have entered is not valid.  Please check the spelling of the account name and password.  If you need help in retrieving a lost or stolen password and account
    CE_ACCOUNT_IN_USE               = $06; //This account is already logged in.  Please check the spelling and try again.
    CE_PREORDER_TIME_LIMIT          = $07;
	  CE_SERVER_FULL                  = $08; //Could not log in at this time.  Please try again later.
	  CE_WRONG_BUILD_NUMBER           = $09; //Unable to validate game version.  This may be caused by file corruption or the interference of another program.
	  CE_UPDATE_CLIENT                = $0A;
    CE_ACCOUNT_FREEZED              = $0C;

    CMD_AUTH_LOGON_CHALLENGE        = $00;
    CMD_AUTH_LOGON_PROOF            = $01;
    CMD_AUTH_RECONNECT_CHALLENGE    = $02;
    CMD_AUTH_RECONNECT_PROOF        = $03;
    //CMD_update srv =4
    CMD_REALM_LIST                  = $10;
    CMD_XFER_INITIATE               = $30;
    CMD_XFER_DATA                   = $31;
    CMD_XFER_ACCEPT                 = $32;
    CMD_XFER_RESUME                 = $33;
    CMD_XFER_CANCEL                 = $34;

    salt_template: array[0..31]of byte = ($63, $82, $25, $B9, $4D, $FA, $B3, $C2,
       $05, $AD, $D6, $37, $B5, $4C, $00, $F4,
       $C4, $CF, $F0, $CF, $5C, $F1, $75, $6E,
       $AB, $0E, $42, $B6, $FD, $4A, $30, $4E);


var
  ReadBufBySock: array[0..MAXWORD]of TReadBuffer;
  SendBufBySock: array[0..MAXWORD]of TSendBuffer;
  AuthProcTable : Array [CMD_AUTH_LOGON_CHALLENGE..CMD_XFER_CANCEL] of AuthProcs;
  ThreadedAccountsBySockHT: TThreadedIntegerHashtable; //authenticated players account name by socket

{void AuthCrypt::SetKey(BigNumber *bn)

    uint8 *key = new uint8[SHA_DIGEST_LENGTH];
    GenerateKey(key, bn);
    _key.resize(SHA_DIGEST_LENGTH);
    std::copy(key, key + SHA_DIGEST_LENGTH, _key.begin());
    delete[] key;
}
 
{void AuthCrypt::GenerateKey(uint8 *key, BigNumber *bn)

    HmacHash hash;
    hash.UpdateBigNumber(bn);
    hash.Finalize();
    memcpy(key, hash.GetDigest(), SHA_DIGEST_LENGTH);
}

{ TAuthPlayer }

constructor TAuthPlayer.Create; //(sock: TSocket)
begin
    //Socket:= sock;
    IsAuthed:= false;

    biN:= BN_new;
    biB:= BN_new;
    biV:= BN_new;
    biChRes:= BN_new;
    BN_hex2bn(biN, PAnsiChar(BIGNUMBER_N_STR));
    BN_rand(biB, SHA_DIGEST_LENGTH * 8, 0, 1);
    //BN_hex2bn(biB,PAnsiChar(BIGNUMBER_B_STR));

    //get from table
    //move(salt_template, salt, sizeof(salt));
    //Seed:=random(maxint);
end;

destructor TAuthPlayer.Destroy;
begin
     BN_free(biB);
     BN_free(biN);
     BN_free(biV);
     BN_free(biChRes);
     inherited;
end;

{ TThreadedAuthPlayer }

constructor TThreadedAuthPlayer.Create; //(sock: TSocket)
{var
   intHT: TIntegerHashtable;
   tempAuthPlayer: TAuthPlayer;}
begin
     inherited Create;
     InitializeCriticalSection(FLock);
     FObject:= TAuthPlayer.Create;
end;

destructor TThreadedAuthPlayer.Destroy;
begin
     Lock;
     try
        FObject.Free;
        inherited Destroy;
     finally
        Unlock;
        DeleteCriticalSection(FLock);
     end;
end;

function TThreadedAuthPlayer.Lock: TAuthPlayer;
begin
     EnterCriticalSection(FLock);
     result:=FObject;
end;

procedure TThreadedAuthPlayer.Unlock;
begin
     LeaveCriticalSection(FLock);
end;

{ handlers }

procedure OnAcceptAuthSocket(ppsContext: PPER_SOCKET_CONTEXT);
begin
     ThreadedAccountsBySockHT.Remove(ppsContext.Socket);
     ReadBufBySock[ppsContext.Socket]:= TReadBuffer.Create;
     ppsContext.pIOReadContext.wsa_buf.len:= ReadBufBySock[ppsContext.Socket].GetBuffFreeSpace;
     ppsContext.pIOReadContext.wsa_buf.buf:= ReadBufBySock[ppsContext.Socket].GetDataPtr;
     SendBufBySock[ppsContext.Socket]:= TSendBuffer.Create(SIZE1K);
end;

procedure OnDisconnectAuthSocket(ppsContext: PPER_SOCKET_CONTEXT);
begin
     ThreadedAccountsBySockHT.Remove(ppsContext.Socket);
     FreeAndNil(ReadBufBySock[ppsContext.Socket]);
     FreeAndNil(SendBufBySock[ppsContext.Socket]);
end;

procedure OnReadAuthSocket(ppsContext: PPER_SOCKET_CONTEXT; count: dword);
var
   cmd: byte;
   tempRLB: TReadBuffer;
   data: pointer;
   done: dword;
begin
     if ppsContext.Socket > high(ReadBufBySock) then exit;
     tempRLB:= ReadBufBySock[ppsContext.Socket];
     if tempRLB = nil then exit;
     if not(tempRLB.IncReceived(count)) then begin
        AuthSock.CloseContextSocket(ppsContext.Socket, ClientIoRead);
        exit;
     end;

     {lmsg:='';lstr:='';
     for i:=0 to pred(received_len) do lmsg:=lmsg+IntToHex(Buf[i],2)+' ';
     for i:=0 to pred(received_len) do if Buf[i]<>0 then lstr:=lstr+char(Buf[i])+' ';
     AddToLog('receive auth socket ='+IntToStr(Sock));
     AddToLog('Msg='+lmsg+' ('+lstr+')'+#$d#$a);}

     data:= tempRLB.GetDataPtr;
     cmd:= TBytes(data)[0]; //ppsContext.pIOReadContext.Buffer[0];
     AddToLog('receive auth opcode = ' + IntToStr(cmd));
     done:= tempRLB.GetDataCount;
     if (cmd < length(AuthProcTable))and(@AuthProcTable[cmd] <> nil) then begin
        done:= AuthProcTable[cmd](data, tempRLB.GetDataCount, ppsContext.Socket);
     end;
     if ReadBufBySock[ppsContext.Socket] = nil then exit; //dced during handle
     if done > 0 then tempRLB.ReleaseBufferCount(done);
     ppsContext.pIOReadContext.wsa_buf.len:= tempRLB.GetBuffFreeSpace;
     ppsContext.pIOReadContext.wsa_buf.buf:= tempRLB.GetDataPtr;
end;

procedure OnWriteAuthSocket(ppsContext: PPER_SOCKET_CONTEXT; count: dword);
begin
     Assert(SendBufBySock[ppsContext.Socket] <> nil,'SendBufBySock[ppsContext.Socket] is nil');
     SendBufBySock[ppsContext.Socket].IncSent(count);
     SendBufBySock[ppsContext.Socket].Send(AuthSock, ppsContext.Socket);
end;

procedure SendLogonError(OpCode, ErrorCode: Byte; sckt: TSocket);
var
  SS: TSendSession;
begin
     SS:= TSendSession.Create(SIZE32);
     SS.InitAuth(OpCode);
     SS.Add(byte(0));
     SS.Add(ErrorCode);
     if SendBufBySock[sckt] <> nil then begin
       SendBufBySock[sckt].AddData(SS.data, SS.Offs);
       SendBufBySock[sckt].Send(AuthSock, sckt);
     end;
     SS.Free;
end;

procedure SendProofError(OpCode, ErrorCode: Byte; sckt: TSocket);
var
  SS: TSendSession;
begin
     SS:= TSendSession.Create(SIZE32);
     SS.InitAuth(OpCode);
     SS.Add(ErrorCode);
     SS.Add(byte(3));
     SS.Add(byte(0));
     if SendBufBySock[sckt] <> nil then begin
       SendBufBySock[sckt].AddData(SS.data, SS.Offs);
       SendBufBySock[sckt].Send(AuthSock, sckt);
     end;
     SS.Free;
end;

procedure RandArray(var InArray; length: integer);
var
   i: integer;
begin
     for i:=0 to pred(length) do PByteArray(@InArray)[i]:=Random(maxbyte);
end;

function StoreUsernamePassHash(const username, password: UTF8String; gm_level: integer): integer;
var
  pq: IPostgresQuery;
  p: TParam;
  biX, biN, biV, biG: pBIGNUM;
  salt: array[0..SIZE32-1]of byte;
  mdValue, mdValue2: TMDValue;
  mdLength, gm_level_temp: dword;
  i: integer;
  username_password: UTF8String;
  tap: TThreadedAuthPlayer;
  ap: TAuthPlayer;
const
    QUERY_CREATEACCOUNT = 'insert into accounts (username, v, s, gm_level) values ($1, $2, $3, $4) returning id';
begin
    result:= 0;

    biX:=BN_new; biN:=BN_new; biV:=BN_new; biG:=BN_new;

    PQFillParamDefaults(p, 4, PG_PARAMTYPE_BINARY, PG_PARAMTYPE_BINARY);

    p.paramValues[0]:= PChar(username);
    p.paramLengths[0]:= length(username);

    iPQ.Lock;
    try
      username_password:= username + ':' + password;
      BN_hex2bn(biN, PAnsiChar(BIGNUMBER_N_STR));
      BN_set_word(biG, 7);
      for i := 0 to high(salt) do
        salt[i]:= Random(MAXBYTE);
      //move(salt_template, salt, sizeof(salt));

      GetSHA1Digest(@mdValue, mdLength, PChar(username_password), length(username_password), nil, 0);
      GetSHA1Digest(@mdValue2, mdLength, @salt, length(salt), @mdValue, SHA_DIGEST_LENGTH);
      BnSetBinary(biX, mdValue2, SHA_DIGEST_LENGTH);
      BnModExp(biV, biG, biX, biN); //v = g.ModExp(x, N);

      //username, v, s, gm_level
      p.paramLengths[0]:= length(username);
      p.paramValues[0]:= PAnsiChar(username);
      p.paramLengths[1]:= BN_num_bytes(biV);
      p.paramValues[1]:= PAnsiChar(biV.d);
      p.paramLengths[2]:= sizeof(salt);
      p.paramValues[2]:= @salt[0];
      p.paramLengths[3]:= sizeof(integer);
      gm_level_temp:= bswap(gm_level);
      p.paramValues[3]:= @gm_level_temp;

      pq:= iPQ.ExecQueryParams(QUERY_CREATEACCOUNT, @p);
      if pq = nil then exit;
      if pq.Status <> PGRES_TUPLES_OK then begin
        AddToLog(iPQ.GetError);
        exit;
      end;
      if pq.RecordCount <= 0 then exit;

      result:= bswap(PInteger(pq.Value[0, 0])^);

      tap:= TThreadedAuthPlayer.Create;
      ap:= tap.Lock;
      try
        ap.id:= result;
        ap.username:= username;
        ap.is_active:= true;
        ap.gm_level:= 0;
        ap.client_version:= 0;
        ap.expires:= 0;
        ap.banned:= false;
        ap.IsAuthed:= false;
        ap.timezone:= 2;
        move(salt, ap.salt, sizeof(salt));
        ap.biV:= BN_dup(biV);
        ThreadedAccountsByNameHT.Add(username, tap);
      finally
        tap.UnLock;
      end;
    finally
      BN_free(biX); BN_free(biN); BN_free(biV); BN_free(biG);
      iPQ.Unlock;
    end;
end;

function LRChallenge(pdata: pointer; count: dword; sckt: TSocket):dword;
var
  i: integer;
  B_array,N_array: array[0..SIZE32-1]of byte;
  Buf: PCMSG_CHALLENGE;
  CurrAuthPlayer: TThreadedAuthPlayer;
  ap: TAuthPlayer;
  biG, bGMOD: pBIGNUM;
  SS: TSendSession;
  temp_username: UTF8String;
const
  INITIAL_OFFS = 4;
begin
    result:= 0;
    if count < INITIAL_OFFS then exit;

    Buf:= PCMSG_CHALLENGE(pdata);
    if count < Buf.Size + INITIAL_OFFS then exit;
    result:= Buf.Size + INITIAL_OFFS;

    if (Buf.Build < LsSettings.MinBuild)or(Buf.Build > LsSettings.MaxBuild) then begin //  
      AddToLog('Incorrect Version: ' + IntToStr(Buf.Build));
      SendLogonError(CMD_AUTH_LOGON_CHALLENGE, CE_WRONG_BUILD_NUMBER, sckt);
      exit;
    end;

    temp_username:= NormalizeString(PChar(@Buf.Login));
    CurrAuthPlayer:= ThreadedAccountsByNameHT[temp_username];
    if CurrAuthPlayer = nil then begin
      AddToLog('account retrieve error: ' + temp_username);
      SendLogonError(CMD_AUTH_LOGON_CHALLENGE, CE_WRONG_BUILD_NUMBER, sckt);
      exit;
    end;
    ThreadedAccountsBySockHT[sckt]:= CurrAuthPlayer;

    ap:= CurrAuthPlayer.Lock;
    try
      ap.timezone:= Buf.TimeZone;
      //ap.Socket:= sckt;
      ap.client_version:= Buf.Build;

      biG:= BN_new; bGMOD:= BN_new;
      try
        //from v stored in table BnModExp(biV,biG,bX,biN); //v = g.ModExp(x, N);

        BN_set_word(biG,7);
        BnModExp(bGMOD, biG, ap.biB, ap.biN); //g.ModExp(b, N);
        //B = ((v * 3) + gmod) % N;
        BN_set_word(ap.biChRes, 3);
        BnMul(ap.biChRes, ap.biV);
        BN_add(ap.biChRes, ap.biChRes, bGMOD);
        BnMod(ap.biChRes, ap.biN);

        BnToBytes(ap.biChRes, B_array, SIZE32);
        BnToBytes(ap.biN, N_array, SIZE32);
        
        SS:= TSendSession.Create(SIZE256);
        SS.InitAuth(CMD_AUTH_LOGON_CHALLENGE);
        SS.Add(byte(0));
        SS.Add(byte(CE_SUCCESS));
        SS.Add(@B_array,SIZE32);
        SS.Add(byte(1));          //g length
        SS.Add(byte(7));          //g
        SS.Add(byte(SIZE32));     //n length
        SS.Add(@N_array, SIZE32); //n
        SS.Add(@ap.salt, SIZE32);
        // 16   +  0    
        for i:=0 to 15 do SS.Add(byte(0)); //seed
        SS.Add(byte(0));          //security
        
        if SendBufBySock[sckt] <> nil then begin
          SendBufBySock[sckt].AddData(SS.data, SS.Offs);
          SendBufBySock[sckt].Send(AuthSock, sckt);
        end;

        SS.Free;
      finally
        BN_free(biG); BN_free(bGMOD);
      end;
    finally
      CurrAuthPlayer.UnLock;
    end;
end;

function LRProof(pdata: pointer; count: dword; sckt: TSocket):dword;
var
  i,offs: integer;
  mdValue: TMDValue;
  mdLength: dword; 
  same: boolean;
  biU, biA, biS: pBIGNUM;
  t, n_array, b_array: array[0..SIZE32-1]of byte;
  NHash, userHash: array[0..SHA_DIGEST_LENGTH-1]of byte;
  t1: array[0..15]of byte;
  vK: array[0..SHA_DIGEST_LENGTH*2-1]of byte;
  temp: TBytes;
  Hash7: array[0..0]of byte;
  Buf: PCMSG_PROOF;
  CurrAuthPlayer: TThreadedAuthPlayer;
  ap: TAuthPlayer;
  SS: TSendSession;
  tempstr: string;
begin
    result:= 0;
    if count < sizeof(TCMSG_PROOF) then exit;
    result:= sizeof(TCMSG_PROOF);

    CurrAuthPlayer:= ThreadedAccountsBySockHT[sckt];
    if CurrAuthPlayer = nil then exit;

    ap:= CurrAuthPlayer.Lock;
    biU:= BN_new; biS:= BN_new; biA:= BN_new;
    try
      Buf:= PCMSG_PROOF(pdata);

      BnToBytes(ap.biChRes, b_array, SIZE32);

      GetSHA1Digest(@mdValue, mdLength, @Buf.A, SIZE32, @b_array, SIZE32);

      //S = (A * (v.ModExp(u, N))).ModExp(b, N);
      BnSetBinary(biU, mdValue, SHA_DIGEST_LENGTH);
      BnModExp(biS, ap.biV, biU, ap.biN);
      BnSetBinary(biA, Buf.A, SIZE32);
      BnMul(biS, biA);
      BnModExp(biS, biS, ap.biB, ap.biN);
      BnToBytes(biS, t, SIZE32);

      for i:=0 to high(t1) do t1[i]:= t[i shl 1];
      GetSHA1Digest(@mdValue, mdLength, @t1, length(t1), nil, 0);

      for i:=0 to SHA_DIGEST_LENGTH-1 do vK[i shl 1]:= mdValue[i];
      for i:=0 to high(t1) do t1[i]:= t[(i shl 1)+1];
      GetSHA1Digest(@mdValue, mdLength, @t1, length(t1), nil, 0);
      for i:=0 to SHA_DIGEST_LENGTH-1 do vK[(i shl 1)+1]:= mdValue[i];

      move(vK, ap.Hash, sizeof(ap.Hash));

      tempstr:= '';
      for i:=0 to length(ap.Hash)-1 do tempstr:= tempstr + IntToHex(ap.Hash[i], 2);
      AddToLog(tempstr);

      BnToBytes(ap.biN, n_array, SIZE32);
      GetSHA1Digest(@mdValue, mdLength, @n_array, SIZE32, nil, 0);

      move(mdValue, NHash, SHA_DIGEST_LENGTH);
      Hash7[0]:= 7;
      GetSHA1Digest(@mdValue, mdLength, @Hash7, length(Hash7), nil, 0);

      for i:=0 to SHA_DIGEST_LENGTH-1 do NHash[i]:=NHash[i] xor mdValue[i];

      GetSHA1Digest(@mdValue, mdLength, PAnsiChar(ap.username), length(ap.username), nil, 0);
      move(mdValue, userHash, SHA_DIGEST_LENGTH);

      setlength(temp,SHA_DIGEST_LENGTH*4 + SIZE32*3);
      offs:=0;
      move(NHash[0], temp[offs], SHA_DIGEST_LENGTH);          inc(offs,SHA_DIGEST_LENGTH);
      move(userHash[0], temp[offs], SHA_DIGEST_LENGTH);       inc(offs,SHA_DIGEST_LENGTH);
      move(ap.salt[0], temp[offs], SIZE32);                   inc(offs,SIZE32);
      move(Buf.A[0], temp[offs], SIZE32);                     inc(offs,SIZE32);
      BnToBytes(ap.biChRes, temp[offs], SIZE32);              inc(offs,SIZE32);
      move(vK[0], temp[offs], SHA_DIGEST_LENGTH*2);

      GetSHA1Digest(@mdValue, mdLength, temp, length(temp), nil, 0);

      same:=true;
      for i:=0 to SHA_DIGEST_LENGTH -1  do
           if Buf.M1[i] <> mdValue[i] then begin
              same:=false;break;
           end;

      if same then begin
        ap.IsAuthed:= true;
        setlength(temp, SIZE32+SHA_DIGEST_LENGTH * 3);
        offs:= 0;
        move(Buf.A[0], temp[offs], SIZE32);                  inc(offs, SIZE32);
        move(mdValue, temp[offs], SHA_DIGEST_LENGTH);        inc(offs, SHA_DIGEST_LENGTH);
        move(vK, temp[offs], SHA_DIGEST_LENGTH * 2);
        GetSHA1Digest(@mdValue, mdLength, temp, length(temp), nil, 0);

        SS:= TSendSession.Create(SIZE32);
        SS.InitAuth(CMD_AUTH_LOGON_PROOF);
        SS.Add(byte(0));
        SS.Add(@mdValue, SHA_DIGEST_LENGTH);
        SS.Add(dword(0));
        SS.Add(dword(0));
        SS.Add(word(0));
        AddToLog('logon proof ' + ap.username);

        if SendBufBySock[sckt] <> nil then begin
           SendBufBySock[sckt].AddData(SS.data, SS.Offs);
           SendBufBySock[sckt].Send(AuthSock, sckt);
        end;
        SS.Free;
      end
      else SendProofError(CMD_AUTH_LOGON_PROOF, CE_ACCOUNT_NO_MATCH, sckt);
    finally
      BN_free(biU); BN_free(biA); BN_free(biS);
      temp:= nil;
      CurrAuthPlayer.UnLock;
    end;
end;

function LRReconnectChallenge(pdata: pointer; count: dword; sckt: TSocket):dword;
{var
   i: integer;
   mdctx: EVP_MD_CTX;
   mdValue: TMDValue;
   mdLength: dword;
   CurrAuthPlayer: TThreadedAuthPlayer;
   ap: TAuthPlayer;}
begin
     result:=count;
     {CurrAuthPlayer:=ThreadedAccountsBySockHT[sckt];
     if CurrAuthPlayer = nil then exit;

     ap:=CurrAuthPlayer.Lock;
     with ap do
     try
        EVP_DigestInit(@mdctx, EVP_md5());
        EVP_DigestUpdate(@mdctx, @ap.SS_Hash.Hash,sizeof(ap.SS_Hash.Hash));
        EVP_DigestFinal(@mdctx, @mdValue, mdLength);
        SS.InitAuth(CMD_AUTH_RECONNECT_CHALLENGE);
        SS.Add(@mdValue,mdLength);
        // 16   +  0    
        for i:=0 to 15 do SS.Add(byte(0));
        SS.Add(byte(0));
        if SendBufBySock[sckt] <> nil then begin
          SendBufBySock[sckt].AddSession(SS);
          SendBufBySock[sckt].Send(AuthSock, sckt);
        end;
        //AuthSock.Send(sckt,@SS.data[0],ap.SS.Offs);
     finally CurrAuthPlayer.UnLock;
     end;}
end;

function LRReconnectProof(pdata: pointer; count: dword; sckt: TSocket):dword;
{var
   CurrAuthPlayer: TThreadedAuthPlayer;
   ap: TAuthPlayer;}
begin
     result:=count; //sizeof(TCMSG_REPROOF);
     {CurrAuthPlayer:=ThreadedAccountsBySockHT[sckt];
     if CurrAuthPlayer = nil then exit;

     ap:=CurrAuthPlayer.Lock;
     try
        if ap.IsAuthed then
           SendProofError(ap,CMD_AUTH_RECONNECT_PROOF,CE_SUCCESS)
        else SendProofError(ap,CMD_AUTH_RECONNECT_PROOF,CE_ACCOUNT_NO_MATCH);
     finally CurrAuthPlayer.UnLock;
     end;}
end;

function LRRealmList(pdata: pointer; count: dword; sckt: TSocket):dword;
var
  SS: TSendSession;
  CurrAuthPlayer: TThreadedAuthPlayer;
  list: TStringList;
  i: integer;
  r: TRealm;
  ap: TAuthPlayer;
  tp: TQueryThread;
  p: PQueryThreadParam;
  //pd: PCMSG_REALMLIST;
begin
    result:= 0;
    if count < sizeof(TCMSG_REALMLIST) then exit;
    result:= sizeof(TCMSG_REALMLIST);
    //pd:= pdata; AddToLog('cmd realmlist param = ' + IntToHex(pd.UNK1, 8));

    CurrAuthPlayer:= ThreadedAccountsBySockHT[sckt];
    if CurrAuthPlayer = nil then exit;

    list:= RealmsList.Lock;
    ap:= CurrAuthPlayer.Lock;
    SS:= TSendSession.Create(SIZE512);
    try
      SS.InitAuth(CMD_REALM_LIST);
      SS.Add(word(43)); //packet size

      SS.Add(integer(0));
      SS.Add(word(list.Count));  //realm count, vary byte/word

      tp:= TQueryThread.Create(list.Count, GetCharsCountThread, R_REALM_GET_CHARSCOUNT);
      for i:= 0 to list.Count - 1 do begin
        r:= TRealm(list.Objects[i]);

        p:= tp.Add(i);
        p.socket:= r.socket;
        setlength(p.custom, sizeof(dword) * 2);
        move(p.query_id, p.custom[0], sizeof(dword));
        move(ap.id, p.custom[sizeof(dword)], sizeof(dword));
      end;
      tp.StartThread;
      tp.DoTerminate;

      for i:= 0 to list.Count -1 do begin
        r:= TRealm(list.Objects[i]);      

        SS.Add(r.r_type);       //icon: 0 - normal, 1 - pvp
        SS.Add(r.locked);       //locked flag
        SS.Add(r.color);        //color: 0 - green 1 - red 2 - gray (offline)
        SS.Add(r.name);
        SS.Add(r.address_port); //RealmInetAddr
        SS.Add(r.population);   //single population 0.0 - low, 1.0 - medium, 2.0 - high, 200.0 - recomended, 400.0 - full
        if tp.ParamByIndex[i].done then
          SS.Add(tp.ParamByIndex[i].value[0])    //chars_count, byte
        else SS.Add(byte(0));
        SS.Add(r.locale);       //time zone, 1 - english
        SS.Add(r.unk_6);
      end;
      tp.Free;
      SS.Add(byte(0));
      SS.Add(byte(0));
      SS.PutToPos(word(SS.Offs - 3), 1); //store packet length
      if SendBufBySock[sckt] <> nil then begin
        SendBufBySock[sckt].AddData(SS.data, SS.Offs);
        SendBufBySock[sckt].Send(AuthSock, sckt);
      end;
    finally 
      SS.Free;
      CurrAuthPlayer.UnLock;
      RealmsList.Unlock;
    end;
end;

procedure InitHandlers;
var
  i: integer;
begin
    ThreadedAccountsBySockHT:= TThreadedIntegerHashtable.Create;

    ZeroMemory(@ReadBufBySock[0], sizeof(ReadBufBySock));
    ZeroMemory(@SendBufBySock[0], sizeof(SendBufBySock));

    for i:= CMD_AUTH_LOGON_CHALLENGE to length(AuthProcTable)-1 do
       @AuthProcTable[i]:= nil;

    @AuthProcTable[CMD_AUTH_LOGON_CHALLENGE] := @LRChallenge;
    @AuthProcTable[CMD_AUTH_LOGON_PROOF]     := @LRProof;
    @AuthProcTable[CMD_AUTH_RECONNECT_CHALLENGE] := @LRReconnectChallenge;
    @AuthProcTable[CMD_AUTH_RECONNECT_PROOF] := @LRReconnectProof;
    @AuthProcTable[CMD_REALM_LIST]           := @LRRealmList;
    {@AuthProcTable[CMD_XFER_INITIATE]        := @EmpyOffset;
    @AuthProcTable[CMD_XFER_DATA]            := @EmpyOffset;
    @AuthProcTable[CMD_XFER_ACCEPT]          := @EmpyOffset;
    @AuthProcTable[CMD_XFER_RESUME]          := @EmpyOffset;
    @AuthProcTable[CMD_XFER_CANCEL]          := @EmpyOffset;}
end;

procedure ReleaseBuffers;
var
  i: integer;
begin
    for i:=0 to high(ReadBufBySock) do
      if ReadBufBySock[i]<>nil then
        ReadBufBySock[i].Free;
    for i:=0 to high(SendBufBySock) do
      if SendBufBySock[i]<>nil then
        SendBufBySock[i].Free;
    ThreadedAccountsBySockHT.Free;
end;

initialization

  InitHandlers;
  
finalization

  ReleaseBuffers;

end.
