unit UnitMain;

interface

uses
  Windows, Messages, SysUtils, Variants, Classes, Graphics, Controls, Forms,
  Dialogs, StdCtrls, ExtCtrls, IniFiles, Winsock2, AcedContainers, TypesDefs,
  libpq_fe, PostgresClasses, LogHandler, IOCPHandler, ByteArrayHandler, DateUtils, LibeayFunc,
  ComCtrls;

type
  TLsSettings = record
    LogonServerAddr : string;
    LogonServerPort : string;
    RealmServerAddr : string;
    RealmServerPort : string;
    Host            : string;
    Port            : string;
    User            : string;
    Password        : string;
    LogonDB         : string;
    MinBuild        : dword;
    MaxBuild        : dword;
    Encoding        : string;
    Collate         : string;
  end;

  TForm1 = class(TForm)
    Panel1: TPanel;
    Label1: TLabel;
    Label2: TLabel;
    ButtonCreateAccount: TButton;
    EditUsername: TEdit;
    EditPassword: TEdit;
    Memo1: TMemo;
    ListViewRealms: TListView;
    procedure ButtonCreateAccountClick(Sender: TObject);
    procedure FormCreate(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
  private
    { Private declarations }
  public
    { Public declarations }
    procedure AddShowToLog(const LogString: string);
  end;

  function NormalizeString(const str: string): UTF8String;
  function IsAccountExists(const username: UTF8String): boolean;

var
  Form1: TForm1;
  AppPath, LogFileName: string;
  Form1DestroyPending: boolean;
  iPQ: IPostgres;
  LsSettings: TLsSettings;
  AuthSock, RealmsSock: TListenSocket;
  ThreadedAccountsByNameHT: TThreadedStringHashtable;  //accounts by account name
  RealmsList: TThreadedStringList;

implementation

{$R *.dfm}

uses LoginHandler, RealmsHandler;

function NormalizeString(const str: string): UTF8String;
var
  temp: WideString;
begin
    temp:= WideUpperCase(Utf8Decode(Utf8Encode(str)));
    temp:= Copy(temp, 1, MAX_USERNAME_LEN shl 1);
    result:= Utf8Encode(temp);
end;

function GetFileVersion(const fname: string): String;
{ Helper function to get the actual file version information }
var
   Info: Pointer;
   InfoSize: DWORD;
   FileInfo: PVSFixedFileInfo;
   FileInfoSize: DWORD;
   Tmp: DWORD;
   Major1, Major2, Minor1, Minor2: Integer;
begin
     Result:= '';
     InfoSize := GetFileVersionInfoSize(PChar(fname), Tmp);
     if InfoSize = 0 then raise Exception.Create('Can''t get file version information for '+ fname);
     GetMem(Info, InfoSize);
     try
        GetFileVersionInfo(PChar(fname), 0, InfoSize, Info);
        VerQueryValue(Info, '\', Pointer(FileInfo), FileInfoSize);
        Major1 := FileInfo.dwFileVersionMS shr 16;
        Major2 := FileInfo.dwFileVersionMS and $FFFF;
        Minor1 := FileInfo.dwFileVersionLS shr 16;
        Minor2 := FileInfo.dwFileVersionLS and $FFFF;

        Result := IntToStr(Major1)+'.'+IntToStr(Major2)+'.'+IntToStr(Minor1)+'.'+IntToStr(Minor2);
     finally FreeMem(Info, FileInfoSize);
     end;
end;

procedure ReadConfig;
var
  SettingsIni: TIniFile;
  fn: String;
begin
     fn := AppPath + 'logonserver.ini' ;
     if not FileExists(fn) then begin
        AddToLog('!!! Config not found! Program terminate.');
        Form1.Close;
     end;

     SettingsIni:= TIniFile.Create(fn);
     try
        LsSettings.LogonServerAddr:= SettingsIni.ReadString('LogonServer','LogonServerAddr','localhost');
        LsSettings.LogonServerPort:= SettingsIni.ReadString('LogonServer','LogonServerPort','3724');
        LsSettings.RealmServerAddr:= SettingsIni.ReadString('LogonServer','RealmServerAddr','localhost');
        LsSettings.RealmServerPort:= SettingsIni.ReadString('LogonServer','RealmServerPort','3725');

        LsSettings.MinBuild:= SettingsIni.ReadInteger('Common','MinBuild', 1);
        LsSettings.MaxBuild:= SettingsIni.ReadInteger('Common','MaxBuild', 10000);

        LsSettings.Host:= SettingsIni.ReadString('Postgres','Host','localhost');
        LsSettings.Port:= SettingsIni.ReadString('Postgres','Port','5432');
        LsSettings.User:= SettingsIni.ReadString('Postgres','User','postgres');
        LsSettings.Password:= SettingsIni.ReadString('Postgres','Password','root');
        LsSettings.LogonDB:= SettingsIni.ReadString('Postgres','LogonDB','logon_db');
        LsSettings.Encoding:= SettingsIni.ReadString('Postgres','Encoding','UTF8');
        LsSettings.Collate:= SettingsIni.ReadString('Postgres','Collate','Russian_Russia.1251');
     finally SettingsIni.Free;
     end;
end;

procedure CreateDatabase;
var
  tempPQ: IPostgres;
  pq: IPostgresQuery;
begin
    tempPQ:= TPostgres.Create('', '', '', LsSettings.User, LsSettings.Password);
    if tempPQ.Connect then begin
      pq:= tempPQ.ExecQuery('CREATE DATABASE logon_db ' + #$d#$a +
        'WITH OWNER = ' + LsSettings.User  + #$d#$a +
        'ENCODING = ''' + LsSettings.Encoding + ''''  + #$d#$a +
        'LC_COLLATE = ''' + LsSettings.Collate + ''''  + #$d#$a +
        'LC_CTYPE = ''' + LsSettings.Collate + ''''  + #$d#$a +
        'CONNECTION LIMIT = -1');
      if pq = nil then exit;
      if pq.Status <> PGRES_COMMAND_OK then begin
        AddToLog(tempPQ.GetError);
        exit;
      end;
    end
    else AddToLog(tempPQ.LastErrorString);
end;

procedure LoadAccounts;
var
  pq: IPostgresQuery;
  i, acc_count, len: integer;
  strHT: TStringHashtable;
  tap: TThreadedAuthPlayer;
  ap: TAuthPlayer;
  p: TParam;
  temp_username: UTF8String;
const
    QUERY_GETACCOUNTS = 'select * from accounts';
begin
    PQFillParamDefaults(p, 0, PG_PARAMTYPE_BINARY, PG_PARAMTYPE_BINARY);

    acc_count:= 0;
    iPQ.Lock;
    try
      pq:= iPQ.ExecQueryParams(QUERY_GETACCOUNTS, @p);
      if pq = nil then exit;
      if pq.Status <> PGRES_TUPLES_OK then begin
        AddToLog(iPQ.GetError);
        exit;
      end;

      if pq.RecordCount <= 0 then exit;

      strHT:= ThreadedAccountsByNameHT.Lock;
      try
        for i := 0 to pq.RecordCount - 1 do begin          
          temp_username:= pq.ValueByName[i, 'username'];

          tap:= TThreadedAuthPlayer.Create;
          ap:= tap.Lock;
          try
            ap.id:= bswap(PInteger(pq.ValueByName[i, 'id'])^);

            ap.username:= temp_username;

            ap.is_active:= PInteger(pq.ValueByName[i, 'is_active'])^ <> 0;
            ap.gm_level:= bswap(PInteger(pq.ValueByName[i, 'gm_level'])^);
            ap.expires:= UnixToDateTime(bswap(PInteger(pq.ValueByName[i, 'expires'])^));
            ap.banned:= PInteger(pq.ValueByName[i, 'banned'])^ <> 0;

            len:= pq.ValueLenByName[i, 's'];
            if len > sizeof(ap.salt) then len:= sizeof(ap.salt);
            move(pq.ValueByName[i, 's']^, ap.salt, len);

            len:= pq.ValueLenByName[i, 'v'];
            BnSetBinary(ap.biV, pq.ValueByName[i, 'v']^, len);
          finally
            tap.UnLock;
          end;

          try
            strHT.Add(temp_username, tap);
            inc(acc_count);
          except tap.Free;
          end;
        end;
      finally ThreadedAccountsByNameHT.UnLock;
      end;
    finally
      iPQ.Unlock;
    end;
    AddToLog('loaded ' + IntToStr(acc_count) + ' accounts');
end;

procedure TForm1.FormCreate(Sender: TObject);
var
  LogsDir: string;
  pq: IPostgresQuery;
  Data: TWSAData;
  strHT: TStringHashtable;
const
  CREATE_LOGON_TABLE = 'CREATE TABLE accounts ' + #$d#$a +
    '(id serial, ' + #$d#$a +
    'username varchar(50), ' + #$d#$a +
    'is_active int default 1 NOT NULL, ' + #$d#$a +
    'gm_level int default 0 NOT NULL, ' + #$d#$a +
    //'logged int default 0 NOT NULL, ' + #$d#$a +
    //'client_version int default 0 NOT NULL, ' + #$d#$a +
    'expires int default 0 NOT NULL, ' + #$d#$a +
    'banned int default 0 NOT NULL, ' + #$d#$a +
    'v bytea, ' + #$d#$a +
    's bytea, ' + #$d#$a +
    //'hash bytea, ' + #$d#$a +
    'UNIQUE (username), ' + #$d#$a +
    'PRIMARY KEY (id)); ' + #$d#$a +
    'CREATE INDEX username_pkey ON accounts (username);';
begin
    Form1DestroyPending:= false;
    AppPath := ExtractFilePath(Application.ExeName);
    LogsDir:= AppPath + 'logs\';
    ForceDirectories(LogsDir);
    LogFileName:= LogsDir + FormatDateTime('yyyymmdd_hhnn',Now) + '.log';
    Caption:= 'Logon server v' + GetFileVersion(Application.ExeName);
    ReadConfig;

    CreateDatabase;

    iPQ:= TPostgres.Create('', '', LsSettings.LogonDB, LsSettings.User, LsSettings.Password);
    if iPQ.Connect then begin
      AddToLog(LsSettings.LogonDB + ' connected');
      pq:= iPQ.ExecQuery(CREATE_LOGON_TABLE);
      if pq <> nil then
        if pq.Status <> PGRES_COMMAND_OK then
          AddToLog(iPQ.GetError);
    end
    else AddToLog(iPQ.LastErrorString);

    ThreadedAccountsByNameHT:= TThreadedStringHashtable.Create(true);
    strHT:= ThreadedAccountsByNameHT.Lock;
    try
      strHT.OwnValues:= true;
    finally ThreadedAccountsByNameHT.UnLock;
    end;

    LoadAccounts;
    RealmsList:= TThreadedStringList.Create;
    
    // 
    WSAStartup($202,Data);
    AuthSock:= TListenSocket.Create(LsSettings.LogonServerAddr, StrToInt(LsSettings.LogonServerPort), OnReadAuthSocket, OnWriteAuthSocket, OnAcceptAuthSocket, OnDisconnectAuthSocket);
    if AuthSock.IsOpen then begin
       AddToLog('AuthSock listen on ' + LsSettings.LogonServerAddr + ':' + LsSettings.LogonServerPort);
    end
    else exit;
    RealmsSock:= TListenSocket.Create(LsSettings.RealmServerAddr, StrToInt(LsSettings.RealmServerPort), OnReadRealmSocket, OnWriteRealmSocket, OnAcceptRealmSocket, OnDisconnectRealmSocket);
    if AuthSock.IsOpen then begin
       AddToLog('AuthSock listen on ' + LsSettings.RealmServerAddr + ':' + LsSettings.RealmServerPort);
    end
    else exit;
end;

procedure TForm1.FormDestroy(Sender: TObject);
var
  list: TStringList;
  i: integer;
  r: TRealm;
begin
    Form1DestroyPending:= true;

    RealmsSock.Free;
    AuthSock.Free;
    WSACleanup;

    list:= RealmsList.Lock;
    try
      for i := 0 to list.Count - 1 do begin
        r:= TRealm(list.Objects[i]);
        r.Free;
      end;
    finally
      RealmsList.UnLock;
    end;

    RealmsList.Free;

    ThreadedAccountsByNameHT.Free;
end;

procedure TForm1.AddShowToLog(const LogString: string);
begin
    {$IFDEF CONS}
    WriteLn(LogString);
    {$ELSE}
    Form1.Memo1.Lines.Add(LogString);
    {$ENDIF}
end;

function IsAccountExists(const username: UTF8String): boolean;
var
  strHT: TStringHashtable;
begin
    strHT:= ThreadedAccountsByNameHT.Lock;
    try
      result:= strHT.Contains(username);
    finally ThreadedAccountsByNameHT.UnLock;
    end;
end;

procedure TForm1.ButtonCreateAccountClick(Sender: TObject);
var
  username, password: UTF8String;
begin
    username:= NormalizeString(EditUsername.Text);
    password:= NormalizeString(EditPassword.Text);

    if IsAccountExists(username) then begin
      AddToLog('account ' + username + ' already exists');
      exit;
    end;
    if StoreUsernamePassHash(username, password, 0) <> 0 then
      AddToLog('Account added');
end;

end.
